#
import numpy as np
import random
import matplotlib
import matplotlib.pyplot as plt
#
# размеры клетчатого мира
WORLD_HEIGHT = 7
WORLD_WIDTH = 10

# сила ветра для колонок
WIND = [0, 0, 0, 1, 1, 1, 2, 2, 1, 0]

# возможные действия
ACTION_UP = 0
ACTION_DOWN = 1
ACTION_LEFT = 2
ACTION_RIGHT = 3
# список действий
ACTIONS4 = [ACTION_UP, ACTION_DOWN, ACTION_LEFT, ACTION_RIGHT]

# вероятность для e-greedy
EPSILON = 0.1
# параметр неопределенности
ALPHA = 0.7
# макс повторов
episode_limit = 300
# награда за каждый шаг step
REWARD = -1.0
total_reward_episode = [0] * episode_limit
# начальная точка
START = [2, 0]
# точка цели
GOAL = [5, 8]

# возвращаем новое состояние s
def step4(state, action):
    i, j = state
    if action == ACTION_UP:
        return [max(i - 1 - WIND[j], 0), j]
    elif action == ACTION_DOWN:
        return [max(min(i + 1 - WIND[j], WORLD_HEIGHT - 1), 0), j]
    elif action == ACTION_LEFT:
        return [max(i - WIND[j], 0), max(j - 1, 0)]
    elif action == ACTION_RIGHT:
        return [max(i - WIND[j], 0), min(j + 1, WORLD_WIDTH - 1)]
    else:
        assert False

# выполним эпизод модели
def episode4(q_value, epc=0):
    # счетчик времени в эпизоде
    time = 0
    # начальное состояние
    state = START

    # выберем действие по epsilon-greedy алгоритму
    if np.random.binomial(1, EPSILON) == 1:
        action = np.random.choice(ACTIONS4)
    else:
        values_ = q_value[state[0], state[1], :]
        action = np.random.choice(
            [action_ for action_, value_ in enumerate(values_) if value_ == np.max(values_)])

    # продолжаем выбирать действия до достижения цели
    while state != GOAL:
        next_state = step4(state, action)
        if np.random.binomial(1, EPSILON) == 1:
            next_action = np.random.choice(ACTIONS4)
        else:
            values_ = q_value[next_state[0], next_state[1], :]
            next_action = np.random.choice(
                [action_ for action_, value_ in enumerate(values_) if value_ == np.max(values_)])

        # пересчет ценности по SARSA
        q_value[state[0], state[1], action] += \
            ALPHA * (REWARD
                     + q_value[next_state[0], next_state[1], next_action]
                     - q_value[state[0], state[1], action])
        state = next_state
        action = next_action

        total_reward_episode[epc] += REWARD
        time += 1
    return time

def run_4_show():
    q_value = np.zeros((WORLD_HEIGHT, WORLD_WIDTH, len(ACTIONS4)))
    #episode_limit = 500
    steps = []
    ep = 0
    # цикл расчета эпизодов модели
    while ep < episode_limit:
        steps.append(episode4(q_value, ep))
        ep += 1

    # отобразим оптимальную политику
    optimal_policy = []
    for i in range(0, WORLD_HEIGHT):
        optimal_policy.append([])
        for j in range(0, WORLD_WIDTH):
            if [i, j] == GOAL:
                # на схеме выделим целевую ячейку +!
                optimal_policy[-1].append('+!')
                continue
            bestAction = np.argmax(q_value[i, j, :])
            if bestAction == ACTION_UP:
                optimal_policy[-1].append('В')
            elif bestAction == ACTION_DOWN:
                optimal_policy[-1].append('Н')
            elif bestAction == ACTION_LEFT:
                optimal_policy[-1].append('Л')
            elif bestAction == ACTION_RIGHT:
                optimal_policy[-1].append('П')
            # на схеме в таблице выделим стартовую ячейку !
            optimal_policy[-1][j] += '!' if ([i, j] == START) else ' '
    #
    print('Оптимальная стратегия:')
    for row in optimal_policy:
        print(row)
    print('\n^Сила^ветра^в^каждой^колонке^\n{}'.format(['^'+str(w) for w in WIND]))

    # рассчитаем кол-во эпизодов с вознаграждением меньше среднего
    avg_reward=sum(total_reward_episode) / episode_limit
    filtered = filter(lambda element: element < avg_reward, total_reward_episode)
    print(f'вознаграждений меньших среднего={round(avg_reward,3)} :: {(len(list(filtered)))}')

    plt.plot(np.arange(1, len(steps) + 1),steps)
    plt.plot(total_reward_episode)
    plt.title('длительность эпизода / вознаграждение')
    plt.xlabel('Эпизоды')
    plt.ylabel('Полное вознаграждение / Время')
    plt.show()

run_4_show()
