#
import numpy as np
import random
import matplotlib
import matplotlib.pyplot as plt
#
# размеры клетчатого мира
WORLD_HEIGHT = 9
WORLD_WIDTH = 10

# сила ветра для колонок
WIND = [0, 0, 0, 1, -1, 2, -2, -1, 1, 0]

# вероятность случайного выбора при обучении
EPSILON = 0.05

# параметр для Sarsa
ALPHA = 0.5

# макс повторов
episode_limit = 500

# награда за каждый шаг step
REWARD = -1.0
total_reward_episode = [0] * episode_limit

# начальная точка
START = [2, 0]
# точка цели
GOAL = [7, 7]

# возможные действия
ACTION_UP = 0
ACTION_DOWN = 1
ACTION_LEFT = 2
ACTION_RIGHT = 3
ACTION_LEFT_UP = 4
ACTION_LEFT_DOWN = 5
ACTION_RIGHT_UP = 6
ACTION_RIGHT_DOWN = 7
ACTION_STOP = 8
# список действий
ACTIONS = [ACTION_UP, ACTION_DOWN, ACTION_LEFT, ACTION_RIGHT, ACTION_LEFT_UP,
           ACTION_LEFT_DOWN, ACTION_RIGHT_UP, ACTION_RIGHT_DOWN, ACTION_STOP]

log_wind = [0] * 5 # список регистрации ветра

# возвращаем новое состояние s
def step9(state, action):
    i, j = state
    if WIND[j] > 0:
       windjr = random.randint(0, WIND[j])
    elif WIND[j] < 0:
       windjr = random.randint(WIND[j], 0)
    else:
       windjr = 0
    log_wind[2+windjr] +=1  # запомним в список регистрации ветра

    def in_boundary(x, y):
        return [max(min(x, WORLD_HEIGHT - 1), 0), max(min(y, WORLD_WIDTH - 1), 0)]

    if action == ACTION_UP:
        return in_boundary(i - 1 - windjr, j)
    elif action == ACTION_DOWN:
        return in_boundary(i + 1 - windjr, j)
    elif action == ACTION_LEFT:
        return in_boundary(i - windjr, j - 1)
    elif action == ACTION_RIGHT:
        return in_boundary(i - windjr, j + 1)
    elif action == ACTION_LEFT_UP:
        return in_boundary(i - 1 - windjr, j - 1)
    elif action == ACTION_LEFT_DOWN:
        return in_boundary(i + 1 - windjr, j - 1)
    elif action == ACTION_RIGHT_UP:
        return in_boundary(i - 1 - windjr, j + 1)
    elif action == ACTION_RIGHT_DOWN:
        return in_boundary(i + 1 - windjr, j + 1)
    elif action == ACTION_STOP:
        return in_boundary(i - windjr, j)
    else:
        assert False

# выполним эпизод модели с расчетом по SARSA
def episode9(q_value, epc=0, alfa=ALPHA, eps=EPSILON):
    # счетчик времени(шагов) в эпизоде
    time = 0
    # начальное состояние
    state = START

    # выберем действие по epsilon-greedy алгоритму
    if np.random.binomial(1, eps) == 1:
        action = np.random.choice(ACTIONS)
    else:
        values_ = q_value[state[0], state[1], :]
        action = np.random.choice(
            [action_ for action_, value_ in enumerate(values_) if value_ == np.max(values_)])

    # продолжаем действия до достижения цели
    while state != GOAL:
        next_state = step9(state, action)
        if np.random.binomial(1, eps) == 1:
            next_action = np.random.choice(ACTIONS)
        else:
            values_ = q_value[next_state[0], next_state[1], :]
            next_action = np.random.choice(
                [action_ for action_, value_ in enumerate(values_) if value_ == np.max(values_)])

        # В Sarsa пересчет Q-значения по текущему и следующему состоянию
        q_value[state[0], state[1], action] += \
            (REWARD
             + q_value[next_state[0], next_state[1], next_action]
             - q_value[state[0], state[1], action]) * alfa
        state = next_state
        action = next_action

        total_reward_episode[epc] +=REWARD
        time += 1
    return time

def run_9_show(alfa=ALPHA, eps=EPSILON):
    global log_wind
    q_value = np.zeros((WORLD_HEIGHT, WORLD_WIDTH, len(ACTIONS)))
    #episode_limit = 500
    steps = []
    log_wind = [0] * 5
    ep = 0
    #
    while ep < episode_limit:
        steps.append(episode9(q_value, ep, alfa, eps))
        ep += 1

    # отобразим оптимальную политику
    optimal_policy = []
    for i in range(0, WORLD_HEIGHT):
        optimal_policy.append([])
        for j in range(0, WORLD_WIDTH):
            if [i, j] == GOAL:
                optimal_policy[-1].append('+!')
                continue
            bestAction = np.argmax(q_value[i, j, :])
            if bestAction == ACTION_UP:
                optimal_policy[-1].append('в ')
            elif bestAction == ACTION_DOWN:
                optimal_policy[-1].append('н ')
            elif bestAction == ACTION_LEFT:
                optimal_policy[-1].append('л ')
            elif bestAction == ACTION_RIGHT:
                optimal_policy[-1].append('п ')
            elif bestAction == ACTION_LEFT_UP:
                optimal_policy[-1].append('лв')
            elif bestAction == ACTION_LEFT_DOWN:
                optimal_policy[-1].append('лн')
            elif bestAction == ACTION_RIGHT_UP:
                optimal_policy[-1].append('пв')
            elif bestAction == ACTION_RIGHT_DOWN:
                optimal_policy[-1].append('пн')
            elif bestAction == ACTION_STOP:
                optimal_policy[-1].append('cт')
            # на схеме в таблице выделим стартовую ячейку ЗАГЛАВНЫМИ
            if [i, j] == START : optimal_policy[-1][j] = optimal_policy[-1][j].upper()
    #
    print('Оптимальная стратегия:')
    for row in optimal_policy:
        print(row)
    print('\n^Макс.сила^ветра^в^каждой^колонке^\n{}'.format([(str(w)) if (w<0) else ('^'+str(w)) for w in WIND]))
    wl = zip([-2,-1,0,1,2],log_wind)
    #print(f'\n -2  -1   0   1   2  _такие были ветра_\n{log_wind}')
    print(f'такие были ветра : {list(wl)}')  #.format(w for w in wl))

    # рассчитаем кол-во эпизодов с вознаграждением меньше среднего
    avg_reward=sum(total_reward_episode) / episode_limit
    filtered = filter(lambda element: element < avg_reward, total_reward_episode)
    print(f'вознаграждений меньших среднего={round(avg_reward,3)} :: {(len(list(filtered)))}')

    plt.plot(np.arange(1, len(steps) + 1),steps)
    plt.plot(total_reward_episode)
    plt.title('длительность эпизода /вознаграждение')
    plt.xlabel('Эпизоды')
    plt.ylabel('Полное вознаграждение / Время')
    plt.show()

#run_9_show()

from collections import Counter

# функция расчета моды с помощью метода Counter.most_common
def my_mode(sample):
    u = Counter(sample)
    return [k for k, v in u.items() if v == u.most_common(1)[0][1]]

# список значений для исследования
alpha_options = [0.5,0.6,0.7,0.8,0.9]
epsilon_options = [0.1, 0.05, 0.01]
result_ = []

def windgridSARSA (alfa=ALPHA, eps=EPSILON):
    q_value = np.zeros((WORLD_HEIGHT, WORLD_WIDTH, len(ACTIONS)))
    steps = []
    ep = 0
    #
    while ep < episode_limit:
        steps.append(episode9(q_value, ep, alfa, eps))
        ep += 1
    return steps

for alpha in alpha_options:
  for epsilon in epsilon_options:
    len_episode = []  # * episode_limit
    total_reward_episode = [0] * episode_limit

    # вызов расчета итогов эпизода
    len_episode = windgridSARSA(alpha, epsilon)

    # расчет средних, макс., мин. характеристик
    avg_rewrd =sum(total_reward_episode) / episode_limit
    avg_steps = sum(len_episode) / episode_limit
    max_rewrd = max(total_reward_episode)
    min_rewrd = min(total_reward_episode)
    mode_rewrd= my_mode(total_reward_episode)
    result_.append([alpha, epsilon, avg_rewrd, avg_steps, max_rewrd, min_rewrd, mode_rewrd])

    print('*',end='')
    #print(f'alpha: {alpha}, epsilon: {epsilon}')
    #print(f'Среднее вознаграждение в {episode_limit} эпизодах: {avg_rewrd}')
    #print(f'Средняя длина {episode_limit} эпизодов: {avg_steps}')

# сортируем итоги по сред.вознаграждению по убыванию
result_.sort(key=(lambda lst: lst[2]), reverse=True)
print(f'\nЛучший результат: {result_[0]}\nХудший результат: {result_[-1]}')

